// Test the pseudo-randomness of the digest.

// Over all 2^32 32-bit message values, accumulate the number of hits on values [0,2^24-1] generated by 
// taking each consecutive group of 3 bytes in the message digest.  (We don't bother to wrap around the 
// end.)  For each cluster of 3 bytes, the number of values for which a particular hit count is achieved
// should be Poisson with mean 256.   We can test each distribution with Chi-square.  Since there are
// 62 results, we must expect some low Chi-square tail probabilities - so the overall summary P-value is
// 1-(1-min(chi-square P-value))^62 - and we would reject at the 5% level if the minimum Chi-square tail
// probability was less than 0.00083.

// Bob Hattersley 07Oct08

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "Waterfall.h"

static unsigned char hits1[62][16777216];
static unsigned int hits2[62][524288];  // Have to accumulate into two arrays and 2 1GB arrays fails ...
#define GET256(i,j) (hits2[i][j/32]&(1<<(j%32)))
#define SET256(i,j) (hits2[i][j/32]) |= 1<<(j%32)

void main () {
	unsigned int in, val;
	BitSequence out[64];
	int i, j, hitcount[512], tot;
	hashState hashstate;

	memset(hits1, 0, sizeof(hits1));
	memset(hits2, 0, sizeof(hits2));

	in = 0;
	do {
		Init(&hashstate, 512);
		Update(&hashstate, (BitSequence *) &in, 32);
		Final(&hashstate, out);

		val = out[0]<<8 | (out[1]<<16);
		for (i=0; i<62; i++) {
			val = (val>>8) | (out[i+2]<<16);
			if (!++hits1[i][val]) { // Value is quite likely to overflow
				if (GET256(i,val)) { printf("Too many hits!\n"); exit(0); }
				SET256(i,val);
			}
		}
			
		if ((in&0xFFFFF) == 0) {
			if ((in&0x3FFFFFF) == 0) printf("\n%8.8x ", in);
			printf(".");
			fflush(stdout);
		}
	} while (++in);

	printf("\n");
	for (i=0; i<62; i++) {
		printf("Bytes %d to %d\n", i, i+2);
		memset(hitcount, 0, sizeof(hitcount));
		for (j=0; j<16777216; j++) {
			val = hits1[i][j];
			if(GET256(i,j)) val += 256;
			hitcount[val]++;
		}

		for (j=0; j<=159; j++) if (hitcount[j]) printf ("Outlier! %d\t%d\n", j, hitcount[j]);
		for (j=367; j<512; j++) if (hitcount[j]) printf ("Outlier! %d\t%d\n", j, hitcount[j]);

		tot = 0;
		for (j=0; j<=180; j++) tot += hitcount[j];
		printf ("Up to 180\t%d\n", tot);

		printf ("181-183\t%d\n", hitcount[181]+hitcount[182]+hitcount[183]);

		for (j=184; j<336; j++) printf ("%d\t%d\n", j, hitcount[j]);

		printf ("336-337\t%d\n", hitcount[336]+hitcount[337]);

		tot = 0;
		for (j=338; j<512; j++) tot += hitcount[j];
		printf ("From 338\t%d\n\n", tot);
	}
}
